# -*- coding: utf-8 -*-
"""
********************************************************************************************
Spreader Application for CTM200

Author: Andrew Craig
Copyright (C) 2018, Cypress Solutions
All rights reserved.
********************************************************************************************
"""
import os
import sys
import binascii


# ------------------------------------------------------------------------------
# Global Variables
# ------------------------------------------------------------------------------
# Debug flag for testing off CTM200.  Will not try to run commands to get/set sysdb data
ON_DEVICE = True

VERSION = "2.6"

# Data files to use
SD_DATA_FILE = "/var/data/reports/scripts/spreader_data.txt"
TMP_DATA_FILE = "/tmp/spreader_data.txt"
TMP_CRC = "/tmp/spreader_data.crc"
TMP_DATA_FILE_BAK = "/tmp/spreader_data.bak.txt"
TMP_CRC_BAK = "/tmp/spreader_data.bak.crc"

REPORT_TRIGGER = 1021  # The report number of the general GPSReport
# Used because we detect blast changes and need to trigger a message if
# a change was detected.

# Used to check on the status of the data and crc files
DATA_NONE = 0  # No data file and no CRC
DATA_HALF = 1  # Data file but no CRC
DATA_FULL = 2  # Data file and CRC files
# ------------------------------------------------------------------------------


def crc32_from_file(filename):
    """ Helper Function to generate the CRC for a file

    :param filename: The name of the file to generate the CRC for
    :return: The CRC of the data file
    """
    buf = open(filename, 'rb').read()
    crc = (binascii.crc32(buf) & 0xFFFFFFFF)
    return "%08X" % crc


# ----------------------------------------------------------------------------------------------------------------------
# Base Spreader Class
# ----------------------------------------------------------------------------------------------------------------------
class Spreader:
    def __init__(self):
        pass

    @staticmethod
    def main_data_state():
        """ The tries to read the state of the main spreader data file and if there is a crc file as well.
        So it will either have a full backup ( ie file and crc ) or half ( file but no CRC )

        :return: DATA_FULL if full back up found
                 DATA_NONE if no backup found
                 DATA_HALF if a file is found but no CRC
        """
        if os.path.exists(TMP_DATA_FILE) is True and os.path.exists(TMP_CRC) is True:
            return DATA_FULL
        elif os.path.exists(TMP_DATA_FILE) is True and os.path.exists(TMP_CRC) is False:
            return DATA_HALF
        elif os.path.exists(TMP_DATA_FILE) is False and os.path.exists(TMP_CRC) is False:
            return DATA_NONE
        elif os.path.exists(TMP_DATA_FILE) is False and os.path.exists(TMP_CRC) is True:
            return DATA_NONE

    @staticmethod
    def backup_data_state():
        """ The tries to read the state of the back up spreader data file and if there is a crc file as well.
        So it will either have a full backup ( ie file and crc ) or half ( file but no CRC )

        :return: DATA_FULL if full back up found
                 DATA_NONE if no backup found
                 DATA_HALF if a file is found but no CRC
        """
        if os.path.exists(TMP_DATA_FILE_BAK) is True and os.path.exists(TMP_CRC_BAK) is True:
            return DATA_FULL
        elif os.path.exists(TMP_DATA_FILE_BAK) is True and os.path.exists(TMP_CRC_BAK) is False:
            return DATA_HALF
        elif os.path.exists(TMP_DATA_FILE_BAK) is False and os.path.exists(TMP_CRC_BAK) is False:
            return DATA_NONE
        elif os.path.exists(TMP_DATA_FILE_BAK) is False and os.path.exists(TMP_CRC_BAK) is True:
            return DATA_NONE

    @staticmethod
    def verify_file(filename, checksum_filename):
        """ Verify the CRC of the spreader data file.

        :param filename: The name of the data file to check
        :param checksum_filename:  The name of the file that has the CRC to use
        :return: True if the CRC in the crc file matches the calculated CRC of the data file.
        """
        print("%s,%s" % (filename, checksum_filename))
        crc_file = open(checksum_filename, "rb")
        crc_data = crc_file.read()
        crc_file.close()
        return crc_data == crc32_from_file(filename)

    @staticmethod
    def get_data_file():
        """ Get the data file for spreader info
        This will check the state of the data and backup files and their CRC and use the best combination of data/crc
        that it can fine.

        :return: the open data file object if one was found
                None if no good data file/crc combination could be found.
        """
        main_data_state = Spreader.main_data_state()
        backup_data_state = Spreader.backup_data_state()

        print ("Main Data")
        print(main_data_state)
        # No data files present at all, use defaults
        if main_data_state == DATA_NONE and backup_data_state == DATA_NONE:
            os.popen("cmd event add SPREADER_NO_DATA_FILES")
            return None

        # If we have full main data file and checksum.  If that fails try the backup.
        if main_data_state == DATA_FULL:
            print("Data State Full")
            print(Spreader.verify_file(TMP_DATA_FILE, TMP_CRC))

            if Spreader.verify_file(TMP_DATA_FILE, TMP_CRC) is True:
                data_file = open(TMP_DATA_FILE, "r")
                return data_file
            # Fall into the backup if failed main checksum
            elif backup_data_state == DATA_FULL:
                os.popen("cmd event add SPREADER_MAIN_CRC_FAILED")
                if Spreader.verify_file(TMP_DATA_FILE_BAK, TMP_CRC_BAK) is True:
                    data_file = open(TMP_DATA_FILE_BAK, "r")
                    print("b")
                    return data_file
                else:
                    # Give up if could not use the backup
                    os.popen("cmd event add SPREADER_BACKUP_CRC_FAILED")
                    return None

        # If we have no main data or only half main data check to see if we have a full backup.
        if (main_data_state == DATA_HALF or main_data_state == DATA_NONE) and backup_data_state == DATA_FULL:
            if Spreader.verify_file(TMP_DATA_FILE_BAK, TMP_CRC_BAK) is True:
                data_file = open(TMP_DATA_FILE_BAK, "r")
                print("c")
                return data_file
            else:
                os.popen("cmd event add SPREADER_BACKUP_CRC_FAILED")
                return None
        elif (main_data_state == DATA_HALF) and (backup_data_state == DATA_NONE):
            data_file = open(TMP_DATA_FILE, "r")
            print("f")
            return data_file


        print("HERE")

    @staticmethod
    def validate_raw(raw_data):
        """ Helper function that will check to see if we get ERROR from the AUX1_RAW and write the data as USR_STR_4

        :param raw_data: The raw data read from the AUX1_RAW
        :return: True if the data was not ERROR
                False if the data read was ERROR
        """
        if raw_data == "ERROR\r\n":
            cmd_error = "cmd sysdb set USR_STR_4 %s" % raw_data
            os.popen(cmd_error)
            return False

        return True

    @staticmethod
    def load_field(data, index):
        """ Helper function to read data from the saved data file
        """
        field = data[index].replace("\n", "").split(",")
        field = map(int, field)
        return field

    @staticmethod
    def create_checkusm_file():
        # Calculate the checksum file
        buf = open(TMP_DATA_FILE, 'rb').read()
        crc = (binascii.crc32(buf) & 0xFFFFFFFF)
        crc_file = open(TMP_CRC, "w")
        crc_file.write("%08X" % crc)
        crc_file.close() 


# ======================================================================================================================
# CompuSpread CS550 class
# ======================================================================================================================
class CS550(Spreader):
    def __init__(self):
        self.has_data = False
        self.data_set_extended = None

    def run(self):
        if ON_DEVICE is True:
            raw_data = os.popen("cmd sysdb get AUX1_RAW").read()
            #raw_data = "AUX1_RAW=0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000\r\nOK\r"
            if Spreader.validate_raw(raw_data) is False:
                print('No Raw Data')
                return
            else:
                cmd_data = "cmd sysdb set USR_STR_4 %s" % raw_data[9:-5]
                os.popen(cmd_data)

        raw_data = binascii.a2b_hex(raw_data[9:-5])
        self.data = []
        for c in raw_data:
            self.data.append(ord(c))

        if self.checksum_validate() is False:
            print('Invalid Checksum')
            return

        # Check to see if using the basic or extended data set
        if len(self.data) == 128:
            self.data_set_extended = True
        elif len(self.data) == 88:
            self.data_set_extended = False

        # Bail out if neither mode is found.
        if self.data_set_extended is None:
            print("Not enough data for parsing [%d]" % len(self.data))
            return

        # If we succeeded in loading the last data block
        if self.load_last_data() is True:
            self.has_data = True
            self.parse()
            self.calculate_derived_fields()
            self.build_json()
            self.save_last_data()

            # Write data back to sysdb user strings for the reporting engine
            cmd_1 = "cmd sysdb set USR_STR_1 %s" % self.material_info
            cmd_2 = "cmd sysdb set USR_STR_2 %s" % self.spreader_info
            cmd_3 = "cmd sysdb set USR_STR_3 %s" % self.total_info
            cmd_4 = "cmd sysdb set USR_FLT_9 %s" % VERSION

            os.popen(cmd_1)
            os.popen(cmd_2)
            os.popen(cmd_3)
            os.popen(cmd_4)
        else:
            os.popen("cmd event add SPREADER_NO_TMP_FILE")


    def checksum_validate(self):
        """ Cacluate the checksum of the incoming data from the controller.
        This will also check to see if the payload is all zeros and return false.

        :return: True if the checksum is valid and data is not all 0
        """
        # Calculate the checksum over the data
        length = len(self.data) - 2  # the last 2 bytes are the checksum in the data
        bytes = length / 2
        checksum = 0

        # Sum up the bytes over the string
        z = 0
        all_zero = True
        for x in range(bytes):
            checksum = (checksum + self.data[z] + self.data[z + 1] * 256) & 0xFFFF
            if self.data[z] != 0 or self.data[z+1] != 0:
                all_zero = False

            z = z + 2


        # If all bytes are zero return false to indicate bad checksum
        # since that could potentially cause dips down in the data.
        if all_zero is True:
            print("All zero payload detected")
            return False

        # Verify that it matches the data from the controller
        data_checksum = self.data[length + 1] * 256 + self.data[length]
        result = (checksum == data_checksum)
        # print ("Calculated Checksum: %x" % checksum)
        # print ("Message Checksum: %x" % data_checksum)
        return result

    def parse(self):
        """ Parse the spreader data.
        Checks to see if we have the extended or base data set
        """
        self.status = 0  # Default is "Spreading"
        self.solid_on = 0
        self.prewet_on = 0
        self.antiice_on = 0
        self.operating_mode = 0

        mode = self.data[4] + self.data[5] << 8
        if ((mode & 0b00100000) >> 4) == 1:
            self.liquidOnly = True

        status = (self.data[2]) + (self.data[3] << 8)

        if (status & 1) == 1:
            self.status = 1  # Blasting
        if (status >> 2) & 1 == 1:
            self.status = 2  # Paused
        if (status >> 3) & 1 == 1:
            self.status = 3  # Reversing

        self.operating_mode = ((status & 0b001100000) >> 5)

        mats = self.data[65]
        self.solid_material = mats & 0b00000011
        self.prewet_material = (mats & 0b00001100) >> 2
        self.antiice_material = (mats & 0b00110000) >> 4

        self.solid_act_rate = self.data[8] + (self.data[9] << 8)
        self.solid_set_rate = self.data[14] + (self.data[15] << 8)

        # Must have actual rate and is not paused
        if self.solid_act_rate > 0 and self.status is not 2:
            self.solid_on = 1

        self.prewet_set_rate = self.data[16] + (self.data[17] << 8)
        self.prewet_act_rate = self.data[10] + (self.data[11] << 8)
        if self.prewet_act_rate > 0:
            self.prewet_on = 1

        # Parsing the basic CS550 data
        if self.data_set_extended is False:
            self.trip_solid = self.data[58] + (self.data[59] << 8)
            self.trip_prewet = self.data[60] + (self.data[61] << 8)

            self.trip_spread_distance = self.data[54] + (self.data[55] << 8)
            self.trip_spread_time = self.data[56] + (self.data[57] << 8)

            self.trip_antiice = 0
            self.antiice_set_rate = 0
            self.antiice_act_rate = 0

        # This is the extended data set
        elif self.data_set_extended is True:
            self.trip_solid = self.data[60] + (self.data[61] << 8)
            self.trip_prewet = self.data[52] + (self.data[53] << 8)
            self.trip_antiice = self.data[54] + (self.data[55] << 8)

            self.trip_spread_distance = self.data[86] + (self.data[87] << 8)
            self.trip_spread_time = self.data[88] + (self.data[89] << 8)

            self.antiice_set_rate = self.data[96] + (self.data[97] << 8)
            self.antiice_act_rate = self.data[98] + (self.data[99] << 8)
            if self.antiice_act_rate > 0:
                self.antiice_on = 1

            self.trip_antiice = self.data[54] + (self.data[55] << 8)

    def set_defaults(self):
        self.last_trip_solid = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.last_trip_prewet = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.last_trip_antiice = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.seasonal_solid = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.seasonal_prewet = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.seasonal_antiice = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.previous_status = 0

    def load_last_data(self):
        """ Try to load in the last data set.  Uses the checksum to verify data.
        If the checksum fails it will try to use the backup data.  If that checksum
        fails it will fall back to use default settings
        """
        data_file = self.get_data_file()
        if data_file is not None:
            stored_data = data_file.readlines()
            self.last_trip_solid = Spreader.load_field(stored_data, 0)
            self.last_trip_prewet = Spreader.load_field(stored_data, 1)
            self.last_trip_antiice = Spreader.load_field(stored_data, 2)
            self.seasonal_solid = Spreader.load_field(stored_data, 3)
            self.seasonal_prewet = Spreader.load_field(stored_data, 4)
            self.seasonal_antiice = Spreader.load_field(stored_data, 5)
            self.previous_status = Spreader.load_field(stored_data, 6)
            data_file.close()

        else:
            # Check to make sure that we don't try to do anything if there is file in /var/data but not yet in
            # tmp
            if os.path.exists(SD_DATA_FILE) is True and os.path.exists(TMP_DATA_FILE) is False:
                #os.popen('cp %s %s' % (SD_DATA_FILE, TMP_DATA_FILE)) 
                print("No Data File")
                return False
            else:
                print("No Data File")
                return False
                #self.set_defaults()

        return True


    def save_last_data(self):
        """ Save the spreader data back to disk.
        """
        self.last_trip_solid[self.solid_material] = self.trip_solid
        self.last_trip_prewet[self.prewet_material] = self.trip_prewet
        self.last_trip_antiice[self.antiice_material] = self.trip_antiice

        # Write the data that we need to tmp to use in future messages
        data_file = open(TMP_DATA_FILE, "w")
        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.last_trip_solid))
        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.last_trip_prewet))
        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.last_trip_antiice))

        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.seasonal_solid))
        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.seasonal_prewet))
        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.seasonal_antiice))
        data_file.write("%d" % int(self.status))
        data_file.close()

        Spreader.create_checkusm_file()

    def calculate_derived_fields(self):
        """  Calculate the values that the spreader does not provide.  For CS550 this is the seasonal values.
        :return:
        """
        solid_delta = self.trip_solid - self.last_trip_solid[self.solid_material]
        #print("Trip: %f" % self.trip_solid)
        #print("Solid Deleta: %f" % solid_delta)

        # If solid delta is less than zero then means the controller probably
        # reset to this is data from a new trip.
        # Add in a -100 check to make sure that it's not a blip
        
        new_seasonal_solid = self.seasonal_solid[self.solid_material]
        if solid_delta < 0 and abs(self.trip_solid) < 500:
            new_seasonal_solid += self.trip_solid
            # this is normal delta.  Add in a 1000 kg check as well
        elif 0 < solid_delta < 500:
            new_seasonal_solid += solid_delta
        if new_seasonal_solid >= self.seasonal_solid[self.solid_material]:
            self.seasonal_solid[self.solid_material] = new_seasonal_solid


        antiice_delta = self.trip_antiice - self.last_trip_antiice[self.antiice_material]
        new_seasonal_antiice = self.seasonal_antiice[self.antiice_material]
        # If delta is less than zero then means the controller probably
        # reset to this is data from a new trip.
        # Add in a -100 check to make sure that it's not a blip
        if antiice_delta < 0 and abs(self.trip_antiice) < 500:
            new_seasonal_antiice += self.trip_antiice
        # this is normal delta.  Add in a 1000 kg check as well
        elif 0 < antiice_delta < 500:
            new_seasonal_antiice += antiice_delta
        
        if new_seasonal_antiice >= self.seasonal_antiice[self.antiice_material]:
            self.seasonal_antiice[self.antiice_material] = new_seasonal_antiice


        prewet_delta = self.trip_prewet - self.last_trip_prewet[self.prewet_material]
        new_seasonal_prewet = self.seasonal_prewet[self.prewet_material]

        # If delta is less than zero then means the controller probably
        # reset to this is data from a new trip.
        # Add in a -100 check to make sure that it's not a blip
        if prewet_delta < 0 and abs(self.trip_prewet) < 500:
            new_seasonal_prewet += self.trip_prewet
        # this is normal delta.  Add in a 1000 kg check as well
        elif 0 < prewet_delta < 500:
            new_seasonal_prewet += prewet_delta
        
        if new_seasonal_prewet >= self.seasonal_prewet[self.prewet_material]:
          self.seasonal_prewet[self.prewet_material] = new_seasonal_prewet


    def blasting_change(self):
        return self.previous_status != self.status


    def build_json(self):
        self.material_info = "'{"
        self.total_info = "'{"
        self.spreader_info = "'{"

        self.spreader_info += "\"M\":\"%d;%d;%d;%d;%d\"^" % (
            self.status, self.solid_on, self.antiice_on, self.prewet_on, self.operating_mode)

        self.material_info += "\"MAT\":\"%d;%d;%d\"^" % (
            self.solid_material, self.prewet_material, self.antiice_material)
        self.material_info += "\"RS\":\"%d;%d\"^" % (self.solid_set_rate, self.solid_act_rate)
        self.material_info += "\"RP\":\"%d;%d\"^" % (self.prewet_set_rate, self.prewet_act_rate)
        self.material_info += "\"RA\":\"%d;%d\"" % (self.antiice_set_rate, self.antiice_act_rate)

        self.total_info += "\"TRIP\":\"%d;%d;%d;%d\"^" % (
            self.trip_solid, self.trip_prewet, self.trip_antiice, self.trip_spread_distance)

        self.total_info += "\"SS\":\"%d;%d\"^" % (self.solid_material, self.seasonal_solid[self.solid_material])
        self.total_info += "\"SP\":\"%d;%d\"^" % (self.prewet_material, self.seasonal_prewet[self.prewet_material])
        self.total_info += "\"SA\":\"%d;%d\"" % (self.antiice_material, self.seasonal_antiice[self.antiice_material])

        self.material_info += "}'"
        self.total_info += "}'"
        self.spreader_info += "}'"


# =======================================================================================================================
# Epoke Spreader Parser
# =======================================================================================================================
class Epoke(Spreader):
    def __init__(self):
        self.has_data = False

    def run(self):
        self.material_info = ""
        self.spreader_info = ""
        self.total_info = ""
    
        if ON_DEVICE is True:
            raw_data = os.popen("cmd sysdb get AUX1_RAW").read()
            #raw_data = "AUX1_RAW=*1*20180009;143444;GE13398;SH-4900 E-S;0.0;0.0;0;0;0;0;0;30;0;27.0;31.6;0;0;0.0;0.0;0.0;12.0;3.0;5;0;99.9;0.890563;-1.990545;2112;99.9;-1.0;-1.0;-1.0;96;0.00;0.00\r\nOK\r"


            # Calgary data 
            #raw_data = "AUX1_RAW=*1*20300820;174120;GE13705;SH-4900 E-S;18.7;10.0;1;0;0;0;0;-10;0;27.0;10.8;0;0;158.0;0.0;0.0;5.0;8.0;0;1;99.9;0.890182;-1.990670;2048;99.9;-1.0;-1.0;-1.0;96;19.64;0.00\r\nOK\r"
            #raw_data = "AUX1_RAW=*1*20300820;174122;GE13705;SH-4900 E-S;0.1;45.0;1;0;0;0;0;-10;0;27.0;10.8;0;0;0.0;0.0;0.0;5.0;8.0;0;1;99.9;0.000000;0.000000;2048;0.0;-1.0;-1.0;-1.0;0;27.00;0.00\r\n\OK\r"

            if Spreader.validate_raw(raw_data) is False:
                print ("Not valid data")
                return
            else:
                # Write incoming data into USR_STR_4
                # Remove the AUX1_RAW= and the trailing terminators
                aux_data = raw_data[9:-5]
                # Remove the *1* because GPSGate won't parse it correctly.
                aux_data = aux_data.replace("*1*", "-1-")
                cmd = "cmd sysdb set USR_STR_4 '%s'" % aux_data
                os.popen(cmd)

        # Chunk the data up and verify that there is enough data
        tokens = aux_data.split(";")

        if len(tokens) == 1:
            print("No Tokens Found")
            return

        # Make sure there is a valid start
        if tokens[0].startswith("-1-") is False and tokens[0].startswith("20") is False:
            print("Not a good starting value for string, rejecting")
            return

        # Valid RS232-1 strings should start with 20 and have at least 28 tokens
        if tokens[0].startswith("20") and len(tokens) < 28:
            print("Not enough tokens[%d] for valid RS232-1 string" % len(tokens))
            return

        # Valid RS232-2 strings should start with *1* and have at least 35 tokens
        if tokens[0].startswith("-1-") and len(tokens) < 35:
            print("Not enough tokens[%d] for valid RS232-2 string" % len(tokens))
            return

        # Load the last data set ( for seasonal values )
        if self.load_last_data() is True:
            self.parse(tokens)

            # Make sure the data fields are sane ( ie no -1 values )
            if self.validate_data() is True:
                self.calculateDerivedFields()
                self.buildJSON()
                self.saveLastData()

                cmd_1 = "cmd sysdb set USR_STR_1 %s" % self.material_info
                cmd_2 = "cmd sysdb set USR_STR_2 %s" % self.spreader_info
                cmd_3 = "cmd sysdb set USR_STR_3 %s" % self.total_info
                cmd_4 = "cmd sysdb set USR_FLT_9 %s" % VERSION

                os.popen(cmd_1)
                os.popen(cmd_2)
                os.popen(cmd_3)
                os.popen(cmd_4)

                # Trigger a message if blasting change was detected
                if self.boosting_change() is True:
                    cmd = "cmd sendreport 1021"
                    #os.popen(cmd)

    def parse(self, tokens):
        """ Parse the Epoke data

        :param tokens: The tokens in the Epoke data string
        :return:
        """
        self.date = tokens[0]
        self.time = tokens[1]
        self.solid_on = self.value(tokens[6])
        self.antiice_on = self.value(tokens[7])
        self.prewet_on = self.value(tokens[8])

        self.dry_set_rate = self.value(tokens[13])
        self.antiice_set_rate = self.value(tokens[14])
        self.prewet_set_rate = self.value(tokens[15])

        self.boosting = self.value(tokens[16])

        self.trip_dry = self.value(tokens[17])
        self.trip_antiice = self.value(tokens[18])
        self.trip_prewet = self.value(tokens[19])

        self.dry_width = self.value(tokens[20])
        self.liquid_width = self.value(tokens[21])

        self.dry_material = int(self.value(tokens[22]))
        self.liquid_material = int(self.value(tokens[23]))

        self.dry_actual_rate = None
        self.liquid_actual_rate = None

        self.dry_actual_rate = 0
        self.antiice_actual_rate = 0

        # This is RS232-2 string so get the extra data
        if (len(tokens) >= 35):
            self.dry_actual_rate = self.value(tokens[33])
            self.antiice_actual_rate = self.value(tokens[34])


    def load_last_data(self):
        """ Try to load in the last data set.  Uses the checksum to verify data.
        If the checksum fails it will try to use the backup data.  If that checksum
        fails it will fall back to use default settings
        """
        data_file = self.get_data_file()
        print("Loaded Data File")
        print(data_file)
        if data_file is not None:
            stored_data = data_file.readlines()
            self.last_trip_dry = Spreader.load_field(stored_data, 0)
            self.last_trip_prewet = Spreader.load_field(stored_data, 1)
            self.last_trip_antiice = Spreader.load_field(stored_data, 2)

            self.seasonal_dry = Spreader.load_field(stored_data, 3)
            print("Loadead Data")
            print(self.seasonal_dry)
            self.seasonal_prewet = Spreader.load_field(stored_data, 4)
            self.seasonal_antiice = Spreader.load_field(stored_data, 5)
            self.previous_boosting = Spreader.load_field(stored_data, 6)
            data_file.close()
        else:
            # Check to make sure that we don't try to do anything if there is file in /var/data but not yet in
            # tmp
            if os.path.exists(SD_DATA_FILE) is True and os.path.exists(TMP_DATA_FILE) is False:
                #os.popen('cp %s %s' % (SD_DATA_FILE, TMP_DATA_FILE))  # Don't copy the file
                print("No data on tmp")
                return False
            else:
                print("Could not load data file")
                return False # TOO should report error

        return True


    def set_defaults(self):
        self.last_trip_dry = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.last_trip_prewet = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.last_trip_antiice = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.seasonal_dry = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.seasonal_prewet = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.seasonal_antiice = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.previous_boosting = [0]

    def saveLastData(self):
        #return
        self.last_trip_dry[self.dry_material] = self.trip_dry
        self.last_trip_prewet[self.liquid_material] = self.trip_prewet
        self.last_trip_antiice[self.liquid_material] = self.trip_antiice

        data_file = open(TMP_DATA_FILE, "w")
        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.last_trip_dry))
        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.last_trip_prewet))
        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.last_trip_antiice))

        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.seasonal_dry))
        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.seasonal_prewet))
        data_file.write("%s\n" % ",".join(str(int(x)) for x in self.seasonal_antiice))
        data_file.write("%d\n" % int(self.boosting))
        data_file.close()
        Spreader.create_checkusm_file()


    def calculateDerivedFields(self):
        # Rates are rate * width.  For example g/m^2 * m
        if self.dry_set_rate != None and self.dry_width != None:
            self.dry_set_rate = self.dry_width * self.dry_set_rate
        if self.dry_actual_rate != None and self.dry_width != None:
            self.dry_actual_rate = self.dry_width * self.dry_actual_rate

        if self.antiice_set_rate != None and self.liquid_width != None:
            self.antiice_set_rate = self.liquid_width * self.antiice_set_rate

        solid_delta = self.trip_dry - self.last_trip_dry[self.dry_material]
   

        new_seasonal_dry = self.seasonal_dry[self.dry_material]

        print("Solid Delta: %f" % solid_delta)
        # If solid delta is less than zero then means the controller probably
        # reset to this is data from a new trip.
        # Add in a -100 check to make sure that it's not a blip
        if (solid_delta < 0):
            new_seasonal_dry += self.trip_dry
            # this is normal delta.  Add in a 1000 kg check as well
        elif (solid_delta > 0 and solid_delta < 1000):
            new_seasonal_dry += solid_delta

        if new_seasonal_dry >= self.seasonal_dry[self.dry_material]:
            self.seasonal_dry[self.dry_material] = new_seasonal_dry


        antiice_delta = self.trip_antiice - self.last_trip_antiice[self.liquid_material]
        new_seasonal_antiice = self.seasonal_antiice[self.liquid_material]
        # If delta is less than zero then means the controller probably
        # reset to this is data from a new trip.
        # Add in a -100 check to make sure that it's not a blip
        if (antiice_delta < 0):
          new_seasonal_antiice += self.trip_antiice

        # this is normal delta.  Add in a 500L check as well
        elif (antiice_delta > 0 and antiice_delta < 500):
          new_seasonal_antiice += antiice_delta
        
        if new_seasonal_antiice >= self.seasonal_antiice[self.liquid_material]:
          self.seasonal_antiice[self.liquid_material] = new_seasonal_antiice



        prewet_delta = self.trip_prewet - self.last_trip_prewet[self.liquid_material]
        new_seasonal_prewet = self.seasonal_prewet[self.liquid_material]

        # If delta is less than zero then means the controller probably
        # reset to this is data from a new trip.
        # Add in a -100 check to make sure that it's not a blip
        if (prewet_delta < 0):
            new_seasonal_prewet += self.trip_prewet

        # this is normal delta.  Add in a 500L check as well
        elif prewet_delta > 0 and prewet_delta < 500:
            new_seasonal_prewet += prewet_delta

        if new_seasonal_prewet >= self.seasonal_prewet[self.liquid_material]:
            self.seasonal_prewet[self.liquid_material] = new_seasonal_prewet


    def validate_data(self):
        """ Validate incoming data to see if any of the data points appear abnormal ( ie values < 0 that should not be < 0 )
        """
        # Validate the widths
        if self.liquid_width < 0 or self.dry_width < 0:
            return False

        # Validate antiice rates
        if self.antiice_set_rate < 0 or self.antiice_actual_rate < 0:
            return False

        # Validate dry rates
        if self.dry_set_rate < 0 or self.dry_actual_rate < 0:
            return False

        # Validate prewet rate
        if self.prewet_set_rate < 0:
            return False

        # Validate trip data
        if self.trip_dry < 0 or self.trip_prewet < 0 or self.trip_antiice < 0:
            return False

        return True

    def buildJSON(self):
        self.material_info = "'{"
        self.total_info = "'{"
        self.spreader_info = "'{"

        self.spreader_info += "\"M\":\"%d;%d;%d;%d\"^" % (self.boosting, self.solid_on, self.antiice_on, self.prewet_on)
        self.spreader_info += "\"DT\":\"%s %s\"" % (self.date, self.time)

        self.material_info += "\"MAT\":\"%d;%d;%d\"^" % (self.dry_material, self.liquid_material, self.liquid_material)
        self.material_info += "\"RS\":\"%d;%d\"^" % (self.dry_set_rate, self.dry_actual_rate)
        self.material_info += "\"RP\":\"%d;%d\"^" % (self.prewet_set_rate, 0)
        self.material_info += "\"RA\":\"%d;%d\"^" % (self.antiice_set_rate, self.antiice_actual_rate)
        self.material_info += "\"W\":\"%d;%d\"" % (self.dry_width, self.liquid_width)

        self.total_info += "\"TRIP\":\"%d;%d;%d\"^" % (self.trip_dry, self.trip_prewet, self.trip_antiice)
        self.total_info += "\"SS\":\"%d;%d\"^" % (self.dry_material, self.seasonal_dry[self.dry_material])
        self.total_info += "\"SP\":\"%d;%d\"^" % (self.liquid_material, self.seasonal_prewet[self.liquid_material])
        self.total_info += "\"SA\":\"%d;%d\"" % (self.liquid_material, self.seasonal_antiice[self.liquid_material])

        self.material_info += "}'"
        self.total_info += "}'"
        self.spreader_info += "}'"
        self.hasData = True

        print(self.spreader_info)
        print(self.material_info)
        print(self.total_info)

    def value(self, token):
        if (token != ""):
            return float(token)
        else:
            return 0;

    def boosting_change(self):
        return self.previous_boosting[0] != self.boosting


# =======================================================================================================================
# Circus Spreader Data
# =======================================================================================================================
class Cirus(Spreader):
    def __init__(self):
        self.has_data = False

    def run(self):
        self.setDefaults()
        self.loadLastData()

        self.hasData = False

        try:
            data = open("/tmp/serial_data.txt", "r").read()        
            self.parse(data)
        except:
            print("SPREADER_NO_SERIAL_DATA")

        self.buildJSON()

        cmd_1 = "cmd sysdb set USR_STR_1 %s" % self.material_info
        cmd_2 = "cmd sysdb set USR_STR_2 %s" % self.spreader_info
        cmd_3 = "cmd sysdb set USR_STR_3 %s" % self.total_info
        cmd_4 = "cmd sysdb set USR_FLT_9 %s" % VERSION

        os.popen(cmd_1)
        os.popen(cmd_2)
        os.popen(cmd_3)
        os.popen(cmd_4)

    def parse(self, data):
            data = data.split("\n")
            for item in data:
                tokens = item.replace("\r", "").replace("\n", "").split(",")

                msg_type = tokens[0]
                if msg_type == 'A':
                    self.automode(tokens)
                elif msg_type == 'B':
                    self.matdispensed(tokens)
                elif msg_type == 'C':
                    self.matselection(tokens)
                elif msg_type == 'D':
                    self.distance(tokens)
                elif msg_type == 'F':
                    self.temp(tokens)
                elif msg_type == 'H':
                    self.heartbeat(tokens)
                elif msg_type == 'J':
                    self.dryTotals(tokens)
                elif msg_type == 'K':
                    self.prewetTotals(tokens)
                elif msg_type == 'L':
                    self.antiiceTotals(tokens)
                elif msg_type == 'O':
                    self.mode(tokens)
                elif msg_type == 'R':
                    self.measuredRate(tokens)
                elif msg_type == "Z":
                    self._zmessage(tokens)


    def setDefaults(self):
        self.currentMode = -1
        self.solid_set_rate = -1
        self.prewet_set_rate = -1
        self.antiice_set_rate = -1
        self.trip_solid = -1
        self.trip_prewet = -1
        self.trip_antiice = -1
        self.solid_mat = -1
        self.prewet_mat = -1
        self.antiice_mat = -1
        self.trip_distance = -1
        self.road_temp = -1
        self.air_temp = -1
        self.firmware = -1
        self.truckid = -1
        self.version = -1
        self.trip_spread_distance = -1
        self.seasonal_solid_material = -1
        self.seasonal_solid_total = -1
        self.seasonal_prewet_material = -1
        self.seasonal_prewet_total = -1
        self.seasonal_antiice_material = -1
        self.seasonal_antiice_total = -1
        self.solid_actual_rate = -1
        self.prewet_actual_rate = -1
        self.antiice_actual_rate = -1
        self.blasting = -1
        self.solid_on = -1
        self.prewet_on = -1
        self.antiice_on = -1
        self.low_battery = -1

        self.previous_seasonal_solid   = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.previous_seasonal_prewet  = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.previous_seasonal_antiice = [0, 0, 0, 0, 0, 0, 0, 0, 0]


    def loadLastData(self):
        try:
            data_file = open(TMP_DATA_FILE, "r")
        except IOError:
            print("No previous data file")
            return False

        if data_file is not None:
            stored_data = data_file.readlines()
            data_file.close()

            for data in stored_data:
                data = data.replace("}", "")
                data = data.replace("{", "")
                data = data.replace("\"", "")
                data = data.replace("'", "")
                tokens = data.split("^");

                idx = 0
                for item in tokens:
                    # print "%s [%d]" % (item, idx)
                    line_data = item.split(":")
                    if len(line_data) > 1:
                        tag = line_data[0]
                        values = line_data[1].split(";")

                        if tag == "MAT":
                            if len(values) == 3:
                                self.solid_mat = int(values[0])
                                self.prewet_mat = int(values[1])
                                self.antiice_mat = int(values[2])
                        if tag == "RS":
                            if len(values) == 2:
                                self.solid_set_rate = float(values[0])
                                self.solid_actual_rate = float(values[1])
                        if tag == "RP":
                            if len(values) == 2:
                                self.prewet_set_rate = float(values[0])
                                self.prewet_actual_rate = float(values[1])
                        if tag == "RA":
                            if len(values) == 2:
                                self.antiice_set_rate = float(values[0])
                                self.antiice_actual_rate = float(values[1])
                        if tag == "TRIP":
                            if len(values) == 4:
                                self.trip_solid = float(values[0])
                                self.trip_prewet = float(values[1])
                                self.trip_antiice = float(values[2])
                                self.trip_spread_distance = float(values[3])

                        if tag == "SS":
                            if len(values) == 2:
                                self.seasonal_solid_material = int(values[0])
                                self.seasonal_solid_total = float(values[1])
                                self.previous_seasonal_solid[self.seasonal_solid_material] = self.seasonal_solid_total

                        if tag == "SP":
                            if len(values) == 2:
                                self.seasonal_prewet_material = int(values[0])
                                self.seasonal_prewet_total = float(values[1])
                                self.previous_seasonal_prewet[self.seasonal_prewet_material] = self.seasonal_prewet_total

                        if tag == "SA":
                            if len(values) == 2:
                                self.seasonal_antiice_material = int(values[0])
                                self.seasonal_antiice_total = float(values[1])
                                self.previous_seasonal_antiice[self.seasonal_antiice_material] = self.seasonal_antiice_total

                        if tag == "M":
                            if len(values) == 4:
                                self.blasting = int(values[0])
                                self.solid_on = int(values[1])
                                self.prewet_on = int(values[2])
                                self.antiice_on = int(values[3])
                        if tag == "I":
                            if len(values) >= 4:
                                self.truckid = values[0]
                                self.firmware = values[1]
                                self.version = values[2]
                                self.currentMode = values[3]

                        if tag == "TP":
                            if len(values) == 2:
                                self.air_temp = float(values[0])
                                self.road_temp = float(values[1])

        return True


    def automode(self, tokens):
        # self.drawtokens(tokens)
        self.solid_set_rate = float(tokens[5])
        self.prewet_set_rate = float(tokens[8])
        self.antiice_set_rate = float(tokens[10])
        self.solid_on = int(tokens[4])
        self.prewet_on = int(tokens[7])
        self.antiice_on = int(tokens[9])
        self.blasting = 0
        if tokens[14] == "BLST":
            self.blasting = 1

    def matdispensed(self, tokens):
        self.trip_solid = float(tokens[3])
        self.trip_prewet = float(tokens[4])
        self.trip_antiice = float(tokens[5])

    def matselection(self, tokens):
        self.solid_mat = int(tokens[3][0])
        self.prewet_mat = int(tokens[4][0])
        self.antiice_mat = int(tokens[5][0])

    def distance(self, tokens):
        self.trip_spread_distance = float(tokens[4])

    def temp(self, tokens):
        self.road_temp = float(tokens[3])
        self.air_temp = float(tokens[4])

    def heartbeat(self, tokens):
        self.firmware = tokens[3]
        self.truckid = tokens[4]
        self.version = tokens[5]

    def dryTotals(self, tokens):
        # self.drawtokens(tokens)
        self.seasonal_solid_material = int(tokens[3][0])
        new_total = float(tokens[5])
        if new_total >= self.previous_seasonal_solid[self.seasonal_solid_material]:
            self.seasonal_solid_total = new_total

    def prewetTotals(self, tokens):
        self.seasonal_prewet_material = int(tokens[3][0])
        # Check to make sure going up
        new_total = float(tokens[5])
        if new_total >= self.previous_seasonal_prewet[self.seasonal_prewet_material]:
            self.seasonal_prewet_total = new_total

    def antiiceTotals(self, tokens):
        self.seasonal_antiice_material = int(tokens[3][0])
        new_total = float(tokens[5])
        # Check to make sure going up
        if new_total >= self.previous_seasonal_antiice[self.seasonal_antiice_material]:
            self.seasonal_antiice_total = new_total

    def mode(self, tokens):
        self.currentMode = tokens[3]

    def measuredRate(self, tokens):
        self.solid_actual_rate = float(tokens[3])
        self.prewet_actual_rate = float(tokens[5])
        self.antiice_actual_rate = float(tokens[6])


    def _zmessage(self,tokens):
        if len(tokens) > 4:
            try:
                msg = int(tokens[3])
                if msg == 11:
                    if tokens[4] == "low battery":
                        self.low_battery = 1
            except ValueError:
                print("Non numeric characters in Z message fields")


    def drawtokens(self, tokens):
        idx = 0
        for item in tokens:
            print
            "[%d] %s" % (idx, item)
            idx += 1


    def buildJSON(self):
        self.material_info = "'{"
        self.total_info = "'{"
        self.spreader_info = "'{"

        self.spreader_info += "\"M\":\"%d;%d;%d;%d\"^" % (self.blasting, self.solid_on, self.antiice_on, self.prewet_on)
        self.spreader_info += "\"I\":\"%s;%s;%s;%s;%d\"^" % (self.truckid, self.firmware, self.version, self.currentMode, self.low_battery)
        self.spreader_info += "\"TP\":\"%01.f;%0.1f\"" % (self.air_temp, self.road_temp)

        self.material_info += "\"MAT\":\"%d;%d;%d\"^" % (self.solid_mat, self.prewet_mat, self.antiice_mat)
        self.material_info += "\"RS\":\"%d;%0.1f\"^" % (self.solid_set_rate, self.solid_actual_rate)
        self.material_info += "\"RP\":\"%d;%0.1f\"^" % (self.prewet_set_rate, self.prewet_actual_rate)
        self.material_info += "\"RA\":\"%d;%0.1f\"" % (self.antiice_set_rate, self.antiice_actual_rate)

        self.total_info += "\"TRIP\":\"%d;%d;%d;%d\"^" % (self.trip_solid, self.trip_prewet, self.trip_antiice, self.trip_spread_distance)
        self.total_info += "\"SS\":\"%d;%d\"^" % (self.seasonal_solid_material, self.seasonal_solid_total)
        self.total_info += "\"SP\":\"%d;%d\"^" % (self.seasonal_prewet_material, self.seasonal_prewet_total)
        self.total_info += "\"SA\":\"%d;%d\"" % (self.seasonal_antiice_material, self.seasonal_antiice_total)

        self.material_info += "}'"
        self.total_info += "}'"
        self.spreader_info += "}'"

        data_file = open(TMP_DATA_FILE, "w")
        data_file.write("%s\n" % self.material_info)
        data_file.write("%s\n" % self.total_info)
        data_file.write("%s\n" % self.spreader_info)
        data_file.close()


def main(spreader):
    device = None

    # Construct Spreader Class required and run the parser for data
    if spreader == "CS550":
        device = CS550()
    elif spreader == "EPOKE":
        device = Epoke()
    elif spreader == "CIRUS":
        device = Cirus()

    if device is not None:
        device.run()


if __name__ == "__main__":
    if len(sys.argv) != 2:
        os.popen("cmd event add SPREADER_TYPE_NOT_FOUND")
        print("usage python spreader.py <spreader type>  [ CS550 | EPOKE | CIRUS ]")
    else:
        main(sys.argv[1])



